@@ -21,6 +21,9 @@ Command-line interface to the Neutron APIs
 from __future__ import print_function
 
 import argparse
+import getpass
+import inspect
+import itertools
 import logging
 import os
 import sys
@@ -30,14 +33,19 @@ from keystoneclient.auth.identity import v3 as v3_auth
 from keystoneclient import discover
 from keystoneclient.openstack.common.apiclient import exceptions as ks_exc
 from keystoneclient import session
+from oslo.utils import encodeutils
 import six.moves.urllib.parse as urlparse
 
 from cliff import app
 from cliff import commandmanager
 
+from neutronclient.common import auth_plugin
 from neutronclient.common import clientmanager
+from neutronclient.common import command as openstack_command
 from neutronclient.common import exceptions as exc
+from neutronclient.common import extension as client_extension
 from neutronclient.common import utils
+from neutronclient.i18n import _
 from neutronclient.neutron.v2_0 import agent
 from neutronclient.neutron.v2_0 import agentscheduler
 from neutronclient.neutron.v2_0 import credential
@@ -49,6 +57,11 @@ from neutronclient.neutron.v2_0.fw import firewallrule
 from neutronclient.neutron.v2_0.lb import healthmonitor as lb_healthmonitor
 from neutronclient.neutron.v2_0.lb import member as lb_member
 from neutronclient.neutron.v2_0.lb import pool as lb_pool
+from neutronclient.neutron.v2_0.lb.v2 import healthmonitor as lbaas_healthmon
+from neutronclient.neutron.v2_0.lb.v2 import listener as lbaas_listener
+from neutronclient.neutron.v2_0.lb.v2 import loadbalancer as lbaas_loadbalancer
+from neutronclient.neutron.v2_0.lb.v2 import member as lbaas_member
+from neutronclient.neutron.v2_0.lb.v2 import pool as lbaas_pool
 from neutronclient.neutron.v2_0.lb import vip as lb_vip
 from neutronclient.neutron.v2_0 import metering
 from neutronclient.neutron.v2_0.nec import packetfilter
@@ -68,8 +81,6 @@ from neutronclient.neutron.v2_0.vpn import ikepolicy
 from neutronclient.neutron.v2_0.vpn import ipsec_site_connection
 from neutronclient.neutron.v2_0.vpn import ipsecpolicy
 from neutronclient.neutron.v2_0.vpn import vpnservice
-from neutronclient.openstack.common.gettextutils import _
-from neutronclient.openstack.common import strutils
 from neutronclient.version import __version__
 
 
@@ -115,7 +126,12 @@ def check_non_negative_int(value):
     return value
 
 
+class BashCompletionCommand(openstack_command.OpenStackCommand):
+    """Prints all of the commands and options for bash-completion."""
+    resource = "bash_completion"
+
 COMMAND_V2 = {
+    'bash-completion': BashCompletionCommand,
     'net-list': network.ListNetwork,
     'net-external-list': network.ListExternalNetwork,
     'net-show': network.ShowNetwork,
@@ -163,6 +179,31 @@ COMMAND_V2 = {
     'security-group-rule-show': securitygroup.ShowSecurityGroupRule,
     'security-group-rule-create': securitygroup.CreateSecurityGroupRule,
     'security-group-rule-delete': securitygroup.DeleteSecurityGroupRule,
+    'lbaas-loadbalancer-list': lbaas_loadbalancer.ListLoadBalancer,
+    'lbaas-loadbalancer-show': lbaas_loadbalancer.ShowLoadBalancer,
+    'lbaas-loadbalancer-create': lbaas_loadbalancer.CreateLoadBalancer,
+    'lbaas-loadbalancer-update': lbaas_loadbalancer.UpdateLoadBalancer,
+    'lbaas-loadbalancer-delete': lbaas_loadbalancer.DeleteLoadBalancer,
+    'lbaas-listener-list': lbaas_listener.ListListener,
+    'lbaas-listener-show': lbaas_listener.ShowListener,
+    'lbaas-listener-create': lbaas_listener.CreateListener,
+    'lbaas-listener-update': lbaas_listener.UpdateListener,
+    'lbaas-listener-delete': lbaas_listener.DeleteListener,
+    'lbaas-pool-list': lbaas_pool.ListPool,
+    'lbaas-pool-show': lbaas_pool.ShowPool,
+    'lbaas-pool-create': lbaas_pool.CreatePool,
+    'lbaas-pool-update': lbaas_pool.UpdatePool,
+    'lbaas-pool-delete': lbaas_pool.DeletePool,
+    'lbaas-healthmonitor-list': lbaas_healthmon.ListHealthMonitor,
+    'lbaas-healthmonitor-show': lbaas_healthmon.ShowHealthMonitor,
+    'lbaas-healthmonitor-create': lbaas_healthmon.CreateHealthMonitor,
+    'lbaas-healthmonitor-update': lbaas_healthmon.UpdateHealthMonitor,
+    'lbaas-healthmonitor-delete': lbaas_healthmon.DeleteHealthMonitor,
+    'lbaas-member-list': lbaas_member.ListMember,
+    'lbaas-member-show': lbaas_member.ShowMember,
+    'lbaas-member-create': lbaas_member.CreateMember,
+    'lbaas-member-update': lbaas_member.UpdateMember,
+    'lbaas-member-delete': lbaas_member.DeleteMember,
     'lb-vip-list': lb_vip.ListVip,
     'lb-vip-show': lb_vip.ShowVip,
     'lb-vip-create': lb_vip.CreateVip,
@@ -344,6 +385,11 @@ class NeutronShell(app.App):
         for k, v in self.commands[apiversion].items():
             self.command_manager.add_command(k, v)
 
+        self._register_extensions(VERSION)
+
+        # Pop the 'complete' to correct the outputs of 'neutron help'.
+        self.command_manager.commands.pop('complete')
+
         # This is instantiated in initialize_app() only when using
         # password flow auth
         self.auth_client = None
@@ -435,7 +481,8 @@ class NeutronShell(app.App):
         parser.add_argument(
             '--os-auth-strategy', metavar='<auth-strategy>',
             default=env('OS_AUTH_STRATEGY', default='keystone'),
-            help=_('Valid values are keystone, noauth, and rackspace.'))
+            help=_('DEPRECATED! Only keystone is supported.'))
+
         parser.add_argument(
             '--os_auth_strategy',
             help=argparse.SUPPRESS)
@@ -548,7 +595,7 @@ class NeutronShell(app.App):
             help=_("Path of certificate file to use in SSL "
                    "connection. This file can optionally be "
                    "prepended with the private key. Defaults "
-                   "to env[OS_CERT]"))
+                   "to env[OS_CERT]."))
 
         parser.add_argument(
             '--os-cacert',
@@ -556,7 +603,7 @@ class NeutronShell(app.App):
             default=env('OS_CACERT', default=None),
             help=_("Specify a CA bundle file to use in "
                    "verifying a TLS (https) server certificate. "
-                   "Defaults to env[OS_CACERT]"))
+                   "Defaults to env[OS_CACERT]."))
 
         parser.add_argument(
             '--os-key',
@@ -565,7 +612,7 @@ class NeutronShell(app.App):
             help=_("Path of client key to use in SSL "
                    "connection. This option is not necessary "
                    "if your key is prepended to your certificate "
-                   "file. Defaults to env[OS_KEY]"))
+                   "file. Defaults to env[OS_KEY]."))
 
         parser.add_argument(
             '--os-password', metavar='<auth-password>',
@@ -630,6 +677,25 @@ class NeutronShell(app.App):
                 options.add(option)
         print(' '.join(commands | options))
 
+    def _register_extensions(self, version):
+        for name, module in itertools.chain(
+                client_extension._discover_via_entry_points()):
+            self._extend_shell_commands(module, version)
+
+    def _extend_shell_commands(self, module, version):
+        classes = inspect.getmembers(module, inspect.isclass)
+        for cls_name, cls in classes:
+            if (issubclass(cls, client_extension.NeutronClientExtension) and
+                    hasattr(cls, 'shell_command')):
+                cmd = cls.shell_command
+                if hasattr(cls, 'versions'):
+                    if version not in cls.versions:
+                        continue
+                try:
+                    self.command_manager.add_command(cmd, cls)
+                except TypeError:
+                    pass
+
     def run(self, argv):
         """Equivalent to the main program for the application.
 
@@ -642,7 +708,7 @@ class NeutronShell(app.App):
             help_pos = -1
             help_command_pos = -1
             for arg in argv:
-                if arg == 'bash-completion':
+                if arg == 'bash-completion' and help_command_pos == -1:
                     self._bash_completion()
                     return 0
                 if arg in self.commands[self.api_version]:
@@ -665,26 +731,21 @@ class NeutronShell(app.App):
             self.initialize_app(remainder)
         except Exception as err:
             if self.options.verbose_level >= self.DEBUG_LEVEL:
-                self.log.exception(unicode(err))
+                self.log.exception(err)
                 raise
             else:
-                self.log.error(unicode(err))
+                self.log.error(err)
             return 1
-        result = 1
         if self.interactive_mode:
             _argv = [sys.argv[0]]
             sys.argv = _argv
-            result = self.interact()
-        else:
-            result = self.run_subcommand(remainder)
-        return result
+            return self.interact()
+        return self.run_subcommand(remainder)
 
     def run_subcommand(self, argv):
         subcommand = self.command_manager.find_command(argv)
         cmd_factory, cmd_name, sub_argv = subcommand
         cmd = cmd_factory(self, self.options)
-        err = None
-        result = 1
         try:
             self.prepare_to_run_command(cmd)
             full_name = (cmd_name
@@ -693,35 +754,20 @@ class NeutronShell(app.App):
                          )
             cmd_parser = cmd.get_parser(full_name)
             return run_command(cmd, cmd_parser, sub_argv)
-        except Exception as err:
-            if self.options.verbose_level >= self.DEBUG_LEVEL:
-                self.log.exception(unicode(err))
-            else:
-                self.log.error(unicode(err))
-            try:
-                self.clean_up(cmd, result, err)
-            except Exception as err2:
-                if self.options.verbose_level >= self.DEBUG_LEVEL:
-                    self.log.exception(unicode(err2))
-                else:
-                    self.log.error(_('Could not clean up: %s'), unicode(err2))
+        except Exception as e:
             if self.options.verbose_level >= self.DEBUG_LEVEL:
+                self.log.exception("%s", e)
                 raise
-        else:
-            try:
-                self.clean_up(cmd, result, None)
-            except Exception as err3:
-                if self.options.verbose_level >= self.DEBUG_LEVEL:
-                    self.log.exception(unicode(err3))
-                else:
-                    self.log.error(_('Could not clean up: %s'), unicode(err3))
-        return result
+            self.log.error("%s", e)
+        return 1
 
     def authenticate_user(self):
         """Make sure the user has provided all of the authentication
         info we need.
         """
-        if self.options.os_auth_strategy == 'keystone':
+        _os_auth_str = self.options.os_auth_strategy
+        _auth_plugin = None
+        if _os_auth_str == 'keystone':
             if self.options.os_token or self.options.os_url:
                 # Token flow auth takes priority
                 if not self.options.os_token:
@@ -739,21 +785,32 @@ class NeutronShell(app.App):
                 project_info = (self.options.os_tenant_name or
                                 self.options.os_tenant_id or
                                 (self.options.os_project_name and
-                                    (self.options.project_domain_name or
-                                     self.options.project_domain_id)) or
+                                    (self.options.os_project_domain_name or
+                                     self.options.os_project_domain_id)) or
                                 self.options.os_project_id)
 
                 if (not self.options.os_username
-                    and not self.options.os_user_id):
+                        and not self.options.os_user_id):
                     raise exc.CommandError(
                         _("You must provide a username or user ID via"
                           "  --os-username, env[OS_USERNAME] or"
-                          "  --os-user_id, env[OS_USER_ID]"))
+                          "  --os-user-id, env[OS_USER_ID]"))
 
                 if not self.options.os_password:
-                    raise exc.CommandError(
-                        _("You must provide a password via"
-                          " either --os-password or env[OS_PASSWORD]"))
+                    # No password, If we've got a tty, try prompting for it
+                    if hasattr(sys.stdin, 'isatty') and sys.stdin.isatty():
+                        # Check for Ctl-D
+                        try:
+                            self.options.os_password = getpass.getpass(
+                                'OS Password: ')
+                        except EOFError:
+                            pass
+                    # No password because we didn't have a tty or the
+                    # user Ctl-D when prompted.
+                    if not self.options.os_password:
+                        raise exc.CommandError(
+                            _("You must provide a password via"
+                              " either --os-password or env[OS_PASSWORD]"))
 
                 if (not project_info):
                     # tenent is deprecated in Keystone v3. Use the latest
@@ -773,25 +830,20 @@ class NeutronShell(app.App):
                     raise exc.CommandError(
                         _("You must provide an auth url via"
                           " either --os-auth-url or via env[OS_AUTH_URL]"))
-        elif self.options.os_auth_strategy == 'rackspace':
-            if not self.options.os_token:
-                raise exc.CommandError(
-                    _("You must provide your API Key via"
-                      " either --os-token or env[OS_TOKEN]"))
-
-            if (not self.options.os_username
-                and not self.options.os_user_id):
-                raise exc.CommandError(
-                    _("You must provide a username or user ID via"
-                      "  --os-username, env[OS_USERNAME] or"
-                      "  --os-user_id, env[OS_USER_ID]"))
+            auth_session = self._get_keystone_session()
+            auth = auth_session.auth
         else:   # not keystone
-            if not self.options.os_url:
+            if _os_auth_str:
+                try:
+                    _auth_plugin = auth_plugin.load_plugin(_os_auth_str)
+                except exc.AuthSystemNotFound:
+                    raise
+            elif not self.options.os_url:
                 raise exc.CommandError(
                     _("You must provide a service URL via"
                       " either --os-url or env[OS_URL]"))
-
-        auth_session = self._get_keystone_session()
+            auth_session = None
+            auth = None
 
         self.client_manager = clientmanager.ClientManager(
             token=self.options.os_token,
@@ -805,6 +857,7 @@ class NeutronShell(app.App):
             region_name=self.options.os_region_name,
             api_version=self.api_version,
             auth_strategy=self.options.os_auth_strategy,
+            auth_plugin=_auth_plugin,
             # FIXME (bklei) honor deprecated service_type and
             # endpoint type until they are removed
             service_type=self.options.os_service_type or
@@ -815,10 +868,8 @@ class NeutronShell(app.App):
             timeout=self.options.http_timeout,
             retries=self.options.retries,
             raise_errors=False,
-            # FIXME (roaet) we need to set the session to what it used to be
-            # after the identity issue has been fixed
-            session=None,
-            auth=auth_session.auth,
+            session=auth_session,
+            auth=auth,
             log_credentials=True)
         return
 
@@ -842,11 +893,6 @@ class NeutronShell(app.App):
         if self.interactive_mode or cmd_name != 'help':
             self.authenticate_user()
 
-    def clean_up(self, cmd, result, err):
-        self.log.debug('clean_up %s', cmd.__class__.__name__)
-        if err:
-            self.log.debug('Got an error: %s', unicode(err))
-
     def configure_logging(self):
         """Create logging handlers for any log output."""
         root_logger = logging.getLogger('')
@@ -860,11 +906,18 @@ class NeutronShell(app.App):
                          self.INFO_LEVEL: logging.INFO,
                          self.DEBUG_LEVEL: logging.DEBUG,
                          }.get(self.options.verbose_level, logging.DEBUG)
-        console.setLevel(console_level)
+        # The default log level is INFO, in this situation, set the
+        # log level of the console to WARNING, to avoid displaying
+        # useless messages. This equals using "--quiet"
+        if console_level == logging.INFO:
+            console.setLevel(logging.WARNING)
+        else:
+            console.setLevel(console_level)
         if logging.DEBUG == console_level:
             formatter = logging.Formatter(self.DEBUG_MESSAGE_FORMAT)
         else:
             formatter = logging.Formatter(self.CONSOLE_MESSAGE_FORMAT)
+        logging.getLogger('iso8601.iso8601').setLevel(logging.WARNING)
         logging.getLogger('urllib3.connectionpool').setLevel(logging.WARNING)
         console.setFormatter(formatter)
         root_logger.addHandler(console)
@@ -956,12 +1009,15 @@ class NeutronShell(app.App):
 
 def main(argv=sys.argv[1:]):
     try:
-        return NeutronShell(NEUTRON_API_VERSION).run(map(strutils.safe_decode,
-                                                         argv))
+        return NeutronShell(NEUTRON_API_VERSION).run(
+            list(map(encodeutils.safe_decode, argv)))
+    except KeyboardInterrupt:
+        print("... terminating neutron client", file=sys.stderr)
+        return 130
     except exc.NeutronClientException:
         return 1
     except Exception as e:
-        print(unicode(e))
+        print(e)
         return 1
 
 
